/* Copyright (C) 2012 Kristian Lauszus, TKJ Electronics. All rights reserved.

   This software may be distributed and modified under the terms of the GNU
   General Public License version 2 (GPL2) as published by the Free Software
   Foundation and appearing in the file GPL2.TXT included in the packaging of
   this file. Please note that GPL2 Section 2[b] requires that all works based
   on this software must also be made publicly available under the terms of
   the GPL2 ("Copyleft").

   Contact information
   -------------------

   Kristian Lauszus, TKJ Electronics
Web      :  http://www.tkjelectronics.com
e-mail   :  kristianl@tkjelectronics.com
*/

#ifndef _btd_h_
#define _btd_h_

#include "Usb.h"
#include <stdint.h>
#include "parsetools.h"

// #define ALLOW_SSP 

#define HCI_BUFFERSIZE         256 // max size for ACL data
#define HCI_CMDBUFSIZE         256 
#define L2CAP_IN_BUFFERSIZE    64 

// The implementation of RFCOMM Port Negotiation hard-codes the
// L2CAP output buffer size into the Maximum Frame size LSB.
// This value cannot exceed 269, since a single octet is used to
// encode the length; this was adjusted after SVN #862 to use a
// full 16-bit uint to encode the length 
#define L2CAP_OUT_BUFFERSIZE   127 

// Used in control endpoint header for HCI Commands
#define bmREQ_HCI_OUT USB_SETUP_HOST_TO_DEVICE|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_DEVICE
#define bmREQ_HCI_IN USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_DEVICE
// ACL Data-In pipe read request 
#define bmREQ_ACL_IN  USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_DEVICE

/* Bluetooth HCI states for hci_task() */
#define HCI_INIT_STATE                           0
#define HCI_RESET_STATE                          1
#define HCI_C_TO_H_FLOW_CONTROL_ENABLE           2
#define HCI_READ_HOST_BUFFER_SIZE                3
#define HCI_SET_HOST_BUFFER_SIZE                 4
#define HCI_WRITE_SIMPLE_PAIRING_MODE_STATE      5 
#define HCI_READ_LOCAL_OOB_DATA_STATE            6  
#define HCI_READ_STORED_LINK_KEY                 7  
#define HCI_READ_STORED_LINK_KEYS                8  
#define HCI_WRITE_STORED_LINK_KEY                9  
#define HCI_SET_EVTMASK_STATE                    10
#define HCI_DELETE_STORED_LINK_KEYS              11
#define HCI_READ_LOCAL_SUPPORTED_COMMANDS        12  
#define HCI_READ_LOCAL_SUPPORTED_FEATURES        13
#define HCI_CLEAR_EVENTS_STATE                   14
#define HCI_SET_EVENTS_STATE                     15
#define HCI_WRITE_PIN_TYPE                       16
#define HCI_WRITE_AUTHENTICATION_ENABLE          17
#define HCI_CLASS_STATE                          18
#define HCI_BDADDR_STATE                         19
#define HCI_LOCAL_VERSION_STATE                  20
#define HCI_SET_NAME_STATE                       21
#define HCI_CHECK_WII_SERVICE                    22
#define HCI_INQUIRY_STATE                        23
#define HCI_CONNECT_WII_STATE                    24
#define HCI_CONNECTED_WII_STATE                  25
#define HCI_SCANNING_STATE                       26
#define HCI_WRITE_PAGE_TIMEOUT                   27
#define HCI_WRITE_CONNECTION_ACCEPT_TIMEOUT      28
#define HCI_CONNECT_IN_STATE                     29
#define HCI_ACCEPT_CONNECTION_STATE              30
#define HCI_USER_CONFIRMATION_REQUEST_REPLY      31
#define HCI_REMOTE_NAME_STATE                    32
#define HCI_CONNECT_STATE                        33
#define HCI_HANDLE_RECEIVED                      34
#define HCI_FLOW_CONTROL_ENABLE                  35
#define HCI_CONNECTED_STATE                      36
#define HCI_DISABLE_SCAN_STATE                   37
#define HCI_DONE_STATE                           38
#define HCI_L2CAP_CONFIGURATION_PREAMBLE         39
#define HCI_L2CAP_CONFIGURATION_BEGIN            40
#define HCI_L2CAP_CONFIGURATION_HOLD             41
#define HCI_L2CAP_CONFIGURATION_END              42
#define HCI_ASYNC_HOLD                           43
#define HCI_ACL_U_SETUP                          44
#define HCI_ACL_U_READY                          45
#define HCI_DISCONNECT_STATE                     46
#define HCI_IO_CAPABILITY_REQUESTED_STATE        47
#define HCI_IO_CAPABILITY_RESPONSE_SENT_STATE    48
#define HCI_OOB_DATA_REQUEST_STATE               49
#define HCI_OOB_DATA_REQUEST_REPLY_SENT_STATE    50
#define HCI_SIMPLE_PAIRING_COMPLETE              51
#define HCI_EXTENDED_INQUIRY_RESPONSE_SENT_STATE 52
#define HCI_LINK_KEY_NOTIFICATION_EVENT          53
#define HCI_LINK_KEY_NOTIFICATION_HANDLED        54
#define HCI_CHANGE_CONNECTION_PACKET_TYPE        55
#define HCI_USER_CONFIRMATION_REQUEST            56
#define HCI_CHARACTERIZE_REMOTE                  57
#define HCI_READ_REMOTE_SUPPORTED_FEATURES       58
#define HCI_READ_REMOTE_EXTENDED_FEATURES        59
#define HCI_READ_REMOTE_VERSION_INFORMATION      60
#define HCI_ENCRYPTION_CHANGE                    61

/* HCI event flags*/
#define HCI_FLAG_CMD_COMPLETE                          0x01
#define HCI_FLAG_CONN_COMPLETE                         0x02
#define HCI_FLAG_DISCONN_COMPLETE                      0x04
#define HCI_FLAG_REMOTE_NAME_COMPLETE                  0x08
#define HCI_FLAG_INCOMING_REQUEST                      0x10
#define HCI_FLAG_READ_BDADDR                           0x20
#define HCI_FLAG_READ_VERSION                          0x40
#define HCI_FLAG_NONEMPTY_INQUIRY_RESULT               0x80
#define HCI_FLAG_CONNECT_EVENT                         0x100
#define HCI_FLAG_FLOW_SPECIFICATION                    0x200
#define HCI_FLAG_IO_CAPABILITY_REQUEST                 0x400
#define HCI_FLAG_OOB_DATA_REQUEST_EVENT                0x800
#define HCI_FLAG_SIMPLE_PAIRING_COMPLETE               0x1000
#define HCI_FLAG_LINK_KEY_NOTIFICATION_EVENT           0x2000
#define HCI_FLAG_CHANGE_CONNECTION_PACKET_TYPE         0x4000
#define HCI_FLAG_ENCRYPTION_CHANGE_EVENT               0x8000
#define HCI_FLAG_READ_REMOTE_SUPPORTED_FEATURES_EVENT  0x10000
#define HCI_FLAG_READ_REMOTE_EXTENDED_FEATURES_EVENT   0x20000
#define HCI_FLAG_READ_REMOTE_VERSION_INFORMATION_EVENT 0x40000

/*Macros for HCI event flag tests */
#define hci_cmd_complete (hci_event_flag & HCI_FLAG_CMD_COMPLETE)
#define hci_connect_complete (hci_event_flag & HCI_FLAG_CONN_COMPLETE)
#define hci_disconnect_complete (hci_event_flag & HCI_FLAG_DISCONN_COMPLETE)
#define hci_remote_name_complete (hci_event_flag & HCI_FLAG_REMOTE_NAME_COMPLETE)
#define hci_incoming_connect_request (hci_event_flag & HCI_FLAG_INCOMING_REQUEST)
#define hci_read_bdaddr_complete (hci_event_flag & HCI_FLAG_READ_BDADDR)
#define hci_read_version_complete (hci_event_flag & HCI_FLAG_READ_VERSION)
#define hci_wii_found (hci_event_flag & HCI_FLAG_NONEMPTY_INQUIRY_RESULT)
#define hci_connect_event (hci_event_flag & HCI_FLAG_CONNECT_EVENT)
#define hci_flow_specification_event (hci_event_flag & HCI_FLAG_FLOW_SPECIFICATION)
#define hci_io_capability_request_event (hci_event_flag & HCI_FLAG_IO_CAPABILITY_REQUEST)
#define hci_oob_data_request_event (hci_event_flag & HCI_FLAG_OOB_DATA_REQUEST_EVENT)
#define hci_link_key_notification_event (hci_event_flag & HCI_FLAG_LINK_KEY_NOTIFICATION_EVENT)
#define hci_change_connection_packet_type_event (hci_event_flag & HCI_FLAG_CHANGE_CONNECTION_PACKET_TYPE)
#define hci_encryption_change_event (hci_event_flag & HCI_FLAG_ENCRYPTION_CHANGE_EVENT)
#define hci_read_remote_supported_features_complete_event (hci_event_flag & HCI_FLAG_READ_REMOTE_SUPPORTED_FEATURES_EVENT)
#define hci_read_remote_extended_features_complete_event (hci_event_flag & HCI_FLAG_READ_REMOTE_EXTENDED_FEATURES_EVENT)
#define hci_read_remote_version_information_complete_event (hci_event_flag & HCI_FLAG_READ_REMOTE_VERSION_INFORMATION_EVENT)

/* HCI Events managed */
#define EV_INQUIRY_COMPLETE                             0x01
#define EV_INQUIRY_RESULT                               0x02
#define EV_CONNECT_COMPLETE                             0x03
#define EV_INCOMING_CONNECT                             0x04
#define EV_DISCONNECT_COMPLETE                          0x05
#define EV_AUTHENTICATION_COMPLETE                      0x06
#define EV_REMOTE_NAME_COMPLETE                         0x07
#define EV_ENCRYPTION_CHANGE                            0x08
#define EV_CHANGE_CONNECTION_LINK                       0x09
#define EV_READ_REMOTE_SUPPORTED_FEATURES_COMPLETE      0x0B
#define EV_READ_REMOTE_VERSION_INFORMATION_COMPLETE     0x0C
#define EV_QOS_SETUP_COMPLETE                           0x0D
#define EV_COMMAND_COMPLETE                             0x0E
#define EV_COMMAND_STATUS                               0x0F
#define EV_HARDWARE_ERROR                               0x10
#define EV_ROLE_CHANGED                                 0x12
#define EV_NUM_COMPLETE_PKT                             0x13
#define EV_RETURN_LINK_KEYS                             0x15
#define EV_PIN_CODE_REQUEST                             0x16
#define EV_LINK_KEY_REQUEST                             0x17
#define EV_LINK_KEY_NOTIFICATION                        0x18
#define EV_LOOPBACK_COMMAND                             0x19
#define EV_DATA_BUFFER_OVERFLOW                         0x1A
#define EV_MAX_SLOTS_CHANGE                             0x1B
#define EV_CONNECTION_PACKET_TYPE_CHANGED               0x1D
#define EV_PAGE_SCAN_REP_MODE                           0x20
#define EV_FLOW_SPECIFICATION_COMPLETE                  0x21
#define EV_READ_REMOTE_EXTENDED_FEATURES_COMPLETE       0x23
#define EV_IO_CAPABILITY_REQUEST                        0x31
#define EV_IO_CAPABILITY_REPLY                          0x32
#define EV_USER_CONFIRMATION_REQUEST                    0x33
#define EV_OOB_DATA_REQUEST_EVENT                       0x35
#define EV_SIMPLE_PAIRING_COMPLETE                      0x36
#define EV_USER_PASSKEY_NOTIFICATION                    0x3B
#define EV_MANUFACTURER_HCI_EVENT                       0xFF

/* Event Filter constants (hci_set_event_filter OCF=0x05, OGF=0x03) */
#define EV_FILTER_TYPE_CLEAR_ALL          0x00
#define EV_FILTER_TYPE_INQUIRY_RESULT     0x01
#define EV_FILTER_TYPE_CONNECTION_SETUP   0x02

#define EV_FILTER_COND_INQ_ACCEPT_ALL       0x00
#define EV_FILTER_COND_INQ_CLASS_RESPONDER  0x01
#define EV_FILTER_COND_INQ_ADDR_RESPONDER   0x02
#define EV_FILTER_COND_INQ_CLASS_ANY        0x000000
#define EV_FILTER_COND_INQ_CLASS_ANYMASK    0x000000
#define EV_FILTER_COND_INQ_ADDR_ANY         0x000000000000

#define EV_FILTER_COND_CONN_ACCEPT_ALL       0x00
#define EV_FILTER_COND_CONN_CLASS_MATCHED    0x01
#define EV_FILTER_COND_CONN_ADDR_MATCHED     0x02

#define EV_FILTER_COND_CONN_ALL_NOAUTO          0x01
#define EV_FILTER_COND_CONN_ALL_NO_ROLESWITCH   0x02
#define EV_FILTER_COND_CONN_ALL_WITH_ROLESWITCH 0x03
#define EV_FILTER_COND_CONN_CLASS_ANY           0x000000
#define EV_FILTER_COND_CONN_CLASS_ANYMASK       0x000000
#define EV_FILTER_COND_CONN_ADDR_ANY            0x000000000000

/* HCI Event_Mask */
#define   HCI_Ev_Msk_Inquiry_Complete                             0x0000000000000001
#define   HCI_Ev_Msk_Inquiry_Result                               0x0000000000000002
#define   HCI_Ev_Msk_Connection_Complete                          0x0000000000000004
#define   HCI_Ev_Msk_Connection_Request                           0x0000000000000008
#define   HCI_Ev_Msk_Disconnection_Complete                       0x0000000000000010
#define   HCI_Ev_Msk_Authentication_Complete                      0x0000000000000020
#define   HCI_Ev_Msk_Remote_Name_Request_Complete                 0x0000000000000040
#define   HCI_Ev_Msk_Encryption_Change                            0x0000000000000080
// FF                                                            
#define   HCI_Ev_Msk_Change_Connection_Link_Key_Complete          0x0000000000000100
#define   HCI_Ev_Msk_Master_Link_Key_Complete                     0x0000000000000200
#define   HCI_Ev_Msk_Read_Remote_Supported_Features_Complete      0x0000000000000400
#define   HCI_Ev_Msk_Read_Remote_Version_Information_Complete     0x0000000000000800
#define   HCI_Ev_Msk_QoS_Setup_Complete                           0x0000000000001000
//#define HCI_Ev_Mask_Reserved                                    0x0000000000002000
//#define HCI_Ev_Mask_Reserved                                    0x0000000000004000
#define   HCI_Ev_Msk_Hardware_Error                               0x0000000000008000
// 9F                                                            
#define   HCI_Ev_Msk_Flush_Occurred                               0x0000000000010000
#define   HCI_Ev_Msk_Role_Change                                  0x0000000000020000
//#define HCI_Ev_Mask_Reserved                                    0x0000000000040000
#define   HCI_Ev_Msk_Mode_Change                                  0x0000000000080000
#define   HCI_Ev_Msk_Return_Link_Keys                             0x0000000000100000
#define   HCI_Ev_Msk_PIN_Code_Request                             0x0000000000200000
#define   HCI_Ev_Msk_Link_Key_Request                             0x0000000000400000
#define   HCI_Ev_Msk_Link_Key_Notification                        0x0000000000800000
// FF                                                            
#define   HCI_Ev_Msk_Loopback_Command                             0x0000000001000000
#define   HCI_Ev_Msk_Data_Buffer_Overflow                         0x0000000002000000
#define   HCI_Ev_Msk_Max_Slots_Change                             0x0000000004000000
#define   HCI_Ev_Msk_Read_Clock_Offset_Complete                   0x0000000008000000
#define   HCI_Ev_Msk_Connection_Packet_Type_Changed               0x0000000010000000
#define   HCI_Ev_Msk_QoS_Violation                                0x0000000020000000
#define   HCI_Ev_Msk_Page_Scan_Mode_Change_[deprecated]           0x0000000040000000
#define   HCI_Ev_Msk_Page_Scan_Repetition_Mode_Change             0x0000000080000000
// FF                                                            
#define   HCI_Ev_Msk_Flow_Specification_Complete                  0x0000000100000000
#define   HCI_Ev_Msk_Inquiry_Result_with_RSSI                     0x0000000200000000
#define   HCI_Ev_Msk_Read_Remote_Extended_Features_Complete       0x0000000400000000
//#define HCI_Ev_Mask_Reserved                                    0x0000000800000000
//#define HCI_Ev_Mask_Reserved                                    0x0000001000000000
//#define HCI_Ev_Mask_Reserved                                    0x0000002000000000
//#define HCI_Ev_Mask_Reserved                                    0x0000004000000000
//#define HCI_Ev_Mask_Reserved                                    0x0000008000000000
// 07                                                            
//#define HCI_Ev_Mask_Reserved                                    0x0000010000000000
//#define HCI_Ev_Mask_Reserved                                    0x0000020000000000
//#define HCI_Ev_Mask_Reserved                                    0x0000040000000000
#define   HCI_Ev_Msk_Synchronous_Connection_Complete              0x0000080000000000
#define   HCI_Ev_Msk_Synchronous_Connection_Changed               0x0000100000000000
#define   HCI_Ev_Msk_Sniff_Subrating                              0x0000200000000000
#define   HCI_Ev_Msk_Extended_Inquiry_Result                      0x0000400000000000
#define   HCI_Ev_Msk_Encryption_Key_Refresh_Complete              0x0000800000000000
// F8                                                            
#define   HCI_Ev_Msk_IO_Capability_Request                        0x0001000000000000
#define   HCI_Ev_Msk_IO_Capability_Request_Reply                  0x0002000000000000
#define   HCI_Ev_Msk_User_Confirmation_Request                    0x0004000000000000
#define   HCI_Ev_Msk_User_Passkey_Request                         0x0008000000000000
#define   HCI_Ev_Msk_Remote_OOB_Data_Request                      0x0010000000000000
#define   HCI_Ev_Msk_Simple_Pairing_Complete                      0x0020000000000000
//#define HCI_Ev_Mask_Reserved                                    0x0040000000000000
#define   HCI_Ev_Msk_Link_Supervision_Timeout_Changed             0x0080000000000000
// BF                                                            
#define   HCI_Ev_Msk_Enhanced_Flush_Complete                      0x0100000000000000
//#define HCI_Ev_Mask_Reserved                                    0x0200000000000000
#define   HCI_Ev_Msk_User_Passkey_Notification                    0x0400000000000000
#define   HCI_Ev_Msk_Keypress_Notification                        0x0800000000000000
#define   HCI_Ev_Msk_Remote_Host_Supported_Features_Notification  0x1000000000000000
//#define HCI_Ev_Msk_LE_Meta-Event                                0x2000000000000000
//#define HCI_Ev_Msk_Reserved_for_future_use                      0xC000000000000000
// 1D

/* L2CAP signaling commands */
#define L2CAP_CMD_COMMAND_REJECT                   0x01
#define L2CAP_CMD_CONNECTION_REQUEST               0x02
#define L2CAP_CMD_CONNECTION_RESPONSE              0x03
#define L2CAP_CMD_CONFIG_REQUEST                   0x04
#define L2CAP_CMD_CONFIG_RESPONSE                  0x05
#define L2CAP_CMD_DISCONNECT_REQUEST               0x06
#define L2CAP_CMD_DISCONNECT_RESPONSE              0x07
#define L2CAP_CMD_ECHO_REQUEST                     0x08
#define L2CAP_CMD_ECHO_RESPONSE                    0x09
#define L2CAP_CMD_INFORMATION_REQUEST              0x0A
#define L2CAP_CMD_INFORMATION_RESPONSE             0x0B
#define L2CAP_CMD_CREATE_CHANNEL_REQUEST           0x0C
#define L2CAP_CMD_CREATE_CHANNEL_RESPONSE          0x0D
#define L2CAP_CMD_MOVE_CHANNEL_REQUEST             0x0E
#define L2CAP_CMD_MOVE_CHANNEL_RESPONSE            0x0F
#define L2CAP_CMD_MOVE_CHANNEL_CONFIRMATION        0x10
#define L2CAP_CMD_MOVE_CHANNEL_CONFIRMATION_RESP   0x11
#define L2CAP_CMD_CONNECTION_PARAMETER_UPDATE_RQ   0x12
#define L2CAP_CMD_CONNECTION_PARAMETER_UPDATE_RESP 0x13

// Use this to bootstrap ACL communication from within 
// the state machine
#define L2CAP_CMD_BOOTSTRAP 0xFF

// Used For Connection Response - Remember to Include High Byte
#define PENDING     0x01
#define SUCCESSFUL  0x00

/* Bluetooth L2CAP PSM - see http://www.bluetooth.org/Technical/AssignedNumbers/logical_link.htm */
#define SDP_PSM         0x01 // Service Discovery Protocol PSM Value
#define RFCOMM_PSM      0x03 // RFCOMM PSM Value
#define HID_CTRL_PSM    0x11 // HID_Control PSM Value
#define HID_INTR_PSM    0x13 // HID_Interrupt PSM Value

// Used to determine if it is a Bluetooth dongle
#define WI_SUBCLASS_RF      0x01 // RF Controller
#define WI_PROTOCOL_BT      0x01 // Bluetooth Programming Interface

#define RESET_DEFERRAL_LOOPS 100 // Loop this many times before trying to send the hci reset command
#define BTD_MAX_ENDPOINTS   4
#define BTD_NUMSERVICES     4 // Max number of Bluetooth services - if you need more than four simply increase this number
#define BTD_ENDPOINT_POLLING_INTERVAL 1 // mSec, default 1 millisecond

// DISCARDABLE CONSTANTS
/* Host error result codes, the 4 LSB's in the HRSL register */
#define hrSUCCESS   0x00
#define hrBUSY      0x01
#define hrBADREQ    0x02
#define hrUNDEF     0x03
#define hrNAK       0x04
#define hrSTALL     0x05
#define hrTOGERR    0x06
#define hrWRONGPID  0x07
#define hrBADBC     0x08
#define hrPIDERR    0x09
#define hrPKTERR    0x0A
#define hrCRCERR    0x0B
#define hrKERR      0x0C
#define hrJERR      0x0D
#define hrTIMEOUT   0x0E
#define hrBABBLE    0x0F

/* Bluetooth L2CAP states for SDP_task() */
#define L2CAP_SDP_WAIT                  0
#define L2CAP_SDP_CONFIG_REQUEST        1
#define L2CAP_SDP_REQUEST               2
#define L2CAP_SDP_SUCCESS               3
#define L2CAP_SDP_DONE                  4
#define L2CAP_DISCONNECT_RESPONSE       5

/* Bluetooth L2CAP states for RFCOMM_task() */
#define L2CAP_RFCOMM_WAIT               0
#define L2CAP_RFCOMM_REQUEST            1
#define L2CAP_RFCOMM_SUCCESS            2
#define L2CAP_RFCOMM_DONE               3
#define L2CAP_RFCOMM_HOLD               4

/* L2CAP event flags */
#define L2CAP_FLAG_CONNECTION_SDP_REQUEST       0x001
#define L2CAP_FLAG_CONNECTION_RFCOMM_REQUEST    0x002
#define L2CAP_FLAG_CONFIG_SDP_REQUEST           0x004
#define L2CAP_FLAG_CONFIG_RFCOMM_REQUEST        0x008
#define L2CAP_FLAG_CONFIG_SDP_SUCCESS           0x010
#define L2CAP_FLAG_CONFIG_RFCOMM_SUCCESS        0x020
#define L2CAP_FLAG_DISCONNECT_SDP_REQUEST       0x040
#define L2CAP_FLAG_DISCONNECT_RFCOMM_REQUEST    0x080
#define L2CAP_FLAG_DISCONNECT_RESPONSE          0x100

/* Macros for L2CAP event flag tests */
#define l2cap_connection_request_sdp_flag (l2cap_event_flag & L2CAP_FLAG_CONNECTION_SDP_REQUEST)
#define l2cap_connection_request_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_CONNECTION_RFCOMM_REQUEST)
#define l2cap_config_request_sdp_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_SDP_REQUEST)
#define l2cap_config_request_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_RFCOMM_REQUEST)
#define l2cap_config_success_sdp_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_SDP_SUCCESS)
#define l2cap_config_success_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_RFCOMM_SUCCESS)
#define l2cap_disconnect_request_sdp_flag (l2cap_event_flag & L2CAP_FLAG_DISCONNECT_SDP_REQUEST)
#define l2cap_disconnect_request_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_DISCONNECT_RFCOMM_REQUEST)
#define l2cap_disconnect_response_flag (l2cap_event_flag & L2CAP_FLAG_DISCONNECT_RESPONSE)

/* SDP Protocol Data Unit (PDU) IDs */

#define SDP_ERROR_RESPONSE_PDU                      0x01
#define SDP_SERVICE_SEARCH_REQUEST_PDU              0x02
#define SDP_SERVICE_SEARCH_RESPONSE_PDU             0x03
#define SDP_SERVICE_ATTRIBUTE_REQUEST_PDU           0x04
#define SDP_SERVICE_ATTRIBUTE_RESPONSE_PDU          0x05
#define SDP_SERVICE_SEARCH_ATTRIBUTE_REQUEST_PDU    0x06 // See the RFCOMM specs
#define SDP_SERVICE_SEARCH_ATTRIBUTE_RESPONSE_PDU   0x07 // See the RFCOMM specs
#define SDP_PDU_ID_MAX 7

#define ARDUINO_DUE_PRODUCT_ID 0x003d
#define ARDUINO_DUE_VENDOR_ID 0x2341 
#define L2CAP_UUID          0x0100


#define USB_NORESP_LIMIT    10
#define noresp_limit        ((USB_NORESP_LIMIT > 0 && hci_counter++ > USB_NORESP_LIMIT) ? ((hci_counter = -1), 1) : 0)
#define byteflip(x)  (((x & 0xFF) << 8) | (x >> 8))
#define wordchunk(x,c) (x >> (8 * c)) & 0xFF

#define L2CAP_MESSAGEQUEUE_ITEMS 8 

typedef enum l2cap_queue_item_t {
  L2CAP_QUEUE_NOOP = 0,

  /* Item represents parameters for an L2CAP information response */
  L2CAP_QUEUE_INFORMATION_RESPONSE, // pBtd->l2cap_information_response

  /* Item represents parameters for an L2CAP information request */
  L2CAP_QUEUE_INFORMATION_REQUEST, // pBtd->l2cap_information_request

  L2CAP_SERVICE_NOT_SUPPORTED, // 8 pBtd->serviceNotSupported
  L2CAP_CONNECTION_RESPONSE  , // 9 pBtd->l2cap_connection_response
  L2CAP_CONFIG_REQUEST       , // 10 pBtd->l2cap_config_request
  L2CAP_CONFIG_RESPONSE      , // 11 pBtd->l2cap_config_response
  L2CAP_DISCONNECT_REQUEST   , // 12 pBtd->l2cap_disconnection_request
  L2CAP_DISCONNECT_RESP      , // 13 pBtd->l2cap_disconnection_response

  /* SDP messages */
  L2CAP_SDP_SERVICE_ATTRIBUTE_RESPONSE_PBR, 

  /* RFCOMM messages */
  RFCOMM_PROTOCOL_MESSAGE    ,
  RFCOMM_PROTOCOL_CREDIT     ,

} l2cap_queue_item_t;

typedef enum l2cap_bt_state_change_cbid_t {
  NO_STATE_CHANGE_REQUIRED = 0     ,

  SDP_CONN_INCREMENT_CONNID        ,
  L2CAP_CONFIG_REQUEST_SENT        ,
  L2CAP_CONFIG_RESPONSE_SENT       ,
  SDP_CONN_DISCONNECT_SDP_REQUEST  ,
  SDP_DISCOVERY_MESSAGETOGGLE      ,
  SDP_DISCOVERY_CONNECTIONLESS     ,
  RFCOMM_PROTOCOL_MESSAGE_SENT     ,
  RFCOMM_PROTOCOL_CREDIT_SENT      ,
  RFCOMM_CONN_INCREMENT_CONNID     ,

  RFCOMM_CONN_L2CAP_RFCOMM_REQUEST ,
  RFCOMM_CONN_DISCONNECT_REQUEST   ,
  RFCOMM_CONN_RECEIVED_CONFIG_REPLY,
  RFCOMM_CONN_SEND_COMPLETE
} l2cap_bt_state_change_cbid_t;

// ServiceSearchAttributeResponse structures
// Allows us to construct service descriptions programmatically, 
// defined at compile time, and mutable at runtime.

// Service Discovery Protocol - Service UUIDs
#define SERIALPORT_UUID     0x1101 // See http://www.bluetooth.org/Technical/AssignedNumbers/service_discovery.htm
#define SDP_PublicBrowseRoot 0x1002
#define SDP_PnPInformation   0x1200

struct l2cap_message_t {
  uint16_t handle;           // 0-1
  uint8_t  unused_1[2];      // 2-3
  uint16_t l2cap_length;     // 4-5 ? TODO: CONFIRM / TEST THIS
  uint16_t channel_id;       // 6-7
  uint8_t  msgtype;          // 8
  // 0x01 - Connectionless MTU
  // 0x02 - Extended features supported
  // 0x03 - Fixed channels supported
  uint8_t  identifier;       // 9
  uint8_t  unused_2[2];      // 10-11
  uint16_t psm;              // 12-13
  uint16_t service_class_id; // 14-15 Source service class ID
  uint16_t response;         // 16-17
  uint16_t uuid_high;        // 18-19
} __attribute__((packed));

typedef struct l2cap_message_t l2cap_message_t;

// RFCOMM Frame types (big-endian binary, 'p' indicates P/F bit):
// 000p 1111 (0x0F) - Disconnected Mode
// 001p 1111 (0x2F) - Set Asynchronous Unbalanced Mode (make all control fields 1-octet long)
// 111p 1111 (0xEF) - Unnumbered Information with Header Check
// 000p 0011 (0x03) - Unnumbered Information (not implemented in RFCOMM)
// 010p 0011 (0x43) - Disconnect
// 011p 0011 (0x63) - Unnumbered Acknowlegement
 
#define RFCOMM_DM       0x0F
#define RFCOMM_SABM     0x2F
#define RFCOMM_UIH      0xEF
#define RFCOMM_UI       0x03
#define RFCOMM_DISC     0x43
#define RFCOMM_UA       0x63

#define extendAddress   0x01 // Always 1

// Multiplexer message types
#define BT_RFCOMM_PN_CMD     0x83
#define BT_RFCOMM_PN_RSP     0x81
#define BT_RFCOMM_MSC_CMD    0xE3
#define BT_RFCOMM_MSC_RSP    0xE1
#define BT_RFCOMM_RPN_CMD    0x93
#define BT_RFCOMM_RPN_RSP    0x91
/*
#define BT_RFCOMM_TEST_CMD   0x23
#define BT_RFCOMM_TEST_RSP   0x21
#define BT_RFCOMM_FCON_CMD   0xA3
#define BT_RFCOMM_FCON_RSP   0xA1
#define BT_RFCOMM_FCOFF_CMD  0x63
#define BT_RFCOMM_FCOFF_RSP  0x61
#define BT_RFCOMM_RLS_CMD    0x53
#define BT_RFCOMM_RLS_RSP    0x51
#define BT_RFCOMM_NSC_RSP    0x11
*/

struct l2cap_rfcomm_packet_t {
  uint16_t handle;
  uint8_t unused_1[6];
  uint8_t address; // 8: Channel, direction, Command/Response
  uint8_t control; // 9: Bit 4 is Poll/Final, rest of bits specify control type
  uint8_t length;
  uint8_t command; // 11: Command (per TKJ Electronics code), or length octet 2/2
  uint8_t unused_2[21];
} __attribute__((packed));

typedef struct l2cap_rfcomm_packet_t l2cap_rfcomm_packet_t;

// SDP Protocol Data Unit Header (Core Spec, Vol 3.B.4)
struct sdp_pdu_header {
  uint8_t pdu_id;
  uint16_t transaction_id;
  uint16_t parameter_length;
  uint8_t start_of_data;
} __attribute__((packed));

typedef struct sdp_pdu_header sdp_pdu_header_t;

struct sdp_service_attribute {
  uint16_t attribute_id;
} __attribute__((packed));

struct sdp_data_element_header {
  uint8_t type_descriptor:5;
  uint8_t type_size:3;
} __attribute__((packed));

struct sdp_public_browse_root_response {
  sdp_data_element_header h;
} __attribute__((packed));

// SDP Protocol Attribute Type Descriptors
struct sdp_nil {
  sdp_data_element_header h;
} __attribute__((packed));

#define SDP_NIL 0
#define SDP_UNSIGNED_INT8(m)          ((1<<3)|0),m
#define SDP_UNSIGNED_INT16(m)         ((1<<3)|1),wordchunk(m,1),wordchunk(m,0)
#define SDP_UNSIGNED_INT32(m)         ((1<<3)|2),wordchunk(m,3),wordchunk(m,2),wordchunk(m,1),wordchunk(m,0)
#define SDP_UUID_16(m)                ((3<<3)|1),wordchunk(m,1),wordchunk(m,0)
#define SDP_STRING(n,s...)            ((4<<3)|5),n , ##s
#define SDP_BOOLEAN(m)                ((5<<3)|0),m
#define SDP_DATA_ELEMENT_SEQ_1(m...)  ((6<<3)|5) , ##m
#define SDP_DATA_ELEMENT_SEQ_2(m...)  ((6<<3)|6),wordchunk(m,1),wordchunk(m,0)  
#define SDP_DATA_ELEMENT_SEQ_4(m...)  ((6<<3)|7),wordchunk(m,3),wordchunk(m,2),wordchunk(m,1),wordchunk(m,0)
#define SDP_CODEPOINT_CLOSE(m...)     ((9<<3)|0) , ##m

#define SDP_AttributeList(n,...)      SDP_DATA_ELEMENT_SEQ_1(n),__VA_ARGS__
#define SDP_ServiceRecordHandle(m)    SDP_UNSIGNED_INT16(0x0000),SDP_UNSIGNED_INT32(m)
#define SDP_ServiceClassIDList(n,m...)     SDP_UNSIGNED_INT16(0x0001),SDP_DATA_ELEMENT_SEQ_1(n), ##m
#define SDP_ServiceRecordState(m)     SDP_UNSIGNED_INT16(0x0002),SDP_UNSIGNED_INT32(m)
#define SDP_ServiceID(m)              SDP_UNSIGNED_INT16(0x0003),SDP_UUID_16(m)
#define SDP_ServiceClass(m)           SDP_UUID_16(m)

#define SDP_ProtocolDescriptorList(n,m...) SDP_UNSIGNED_INT16(0x0004),SDP_DATA_ELEMENT_SEQ_1(n) , ##m
#define SDP_Protocol(m,n,q...)        SDP_DATA_ELEMENT_SEQ_1(3+n),SDP_UUID_16(m) , ##q

#define SDP_BrowseGroupList(n,m...)   SDP_UNSIGNED_INT16(0x0005),SDP_DATA_ELEMENT_SEQ_1(n) , ##m
#define SDP_BrowseGroup(m,q...)       SDP_DATA_ELEMENT_SEQ_1(3),SDP_UUID_16(m) , ##q

#define SDP_LanguageBaseAttributeIDList(n,m...) \
                                      SDP_UNSIGNED_INT16(0x0006), \
                                      SDP_DATA_ELEMENT_SEQ_1(n) , ##m
#define SDP_LanguageBaseAttributeTuple(lang, enc, base) \
                                      SDP_UNSIGNED_INT16(lang), \
                                      SDP_UNSIGNED_INT16(enc), \
                                      SDP_UNSIGNED_INT16(base)
#define SDP_ServiceName(base, n, s...) \
                                      SDP_UNSIGNED_INT16(base + 0x0000), \
                                      SDP_STRING(n , ##s)
#define SDP_ServiceDescription(base, n, s...) \
                                      SDP_UNSIGNED_INT16(base + 0x0001), \
                                      SDP_STRING(n , ##s)
                                       
#define SDP_ServiceAvailability(n)    SDP_UNSIGNED_INT16(0x0008),SDP_UNSIGNED_INT8(n)

#define SDP_BluetoothProfileDescriptorList(n,m...) \
                                      SDP_UNSIGNED_INT16(0x0009), \
                                      SDP_DATA_ELEMENT_SEQ_1(n) , ##m
#define SDP_BluetoothProfileElementInt16(uuid,version) \
                                      SDP_DATA_ELEMENT_SEQ_1(0, \
                                        SDP_UUID_16(uuid), \
                                        SDP_UNSIGNED_INT16(version) \
                                      ), SDP_CloseContainer()
#define SDP_BluetoothProfileElementBoolean(uuid,boole) \
                                      SDP_DATA_ELEMENT_SEQ_1(0, \
                                        SDP_UUID_16(uuid), \
                                        SDP_BOOLEAN(boole) \
                                      ), SDP_CloseContainer()

// Use a reserved PDU ID code point to mark recursion termination.
#define SDP_CloseContainer(m...)      SDP_CODEPOINT_CLOSE(##m)

// Command queue transaction items
typedef struct uhd_l2cap_state_information_response_t {
  uint8_t rxid;
  uint8_t infoTypeLow;
  uint8_t infoTypeHigh;
} uhd_l2cap_state_information_response_t;

typedef struct uhd_l2cap_state_information_request_t {
  uint8_t rxid;
  uint8_t infoTypeLow;
  uint8_t infoTypeHigh;
} uhd_l2cap_state_information_request_t;

typedef struct uhd_sdp_response_t {
  uint32_t payload;
  uint16_t service_class_id;
  uint8_t transactionIDHigh;
  uint8_t transactionIDLow;
} uhd_sdp_response_t;

typedef struct uhd_rfcomm_message_out_t {
  uint32_t payload;
  uint16_t scid;
  uint8_t address;
  uint8_t control;
  uint8_t length;
} uhd_rfcomm_message_out_t;

typedef struct uhd_l2cap_connection_response_t {
  uint32_t payload;
  uint8_t rxid;
  uint16_t dcid;
  uint16_t scid;
  uint8_t result;
} uhd_l2cap_connection_response_t;

typedef struct uhd_l2cap_config_request_t {
  uint32_t payload;
  uint8_t rxid;
  uint16_t scid;
} uhd_l2cap_config_request_t;

typedef struct uhd_l2cap_disconnection_message_t {
  uint32_t payload;
  uint8_t rxid;
  uint16_t dcid;
  uint16_t scid;
} uhd_l2cap_disconnection_message_t;

typedef struct uhd_l2cap_queue_t {
  l2cap_queue_item_t type;
  uint32_t id;
  uint16_t handle;
  uint8_t bt_service_index;
  uint8_t queue_index;
  union {
    uhd_l2cap_state_information_response_t re;
    uhd_l2cap_state_information_request_t rq;
    uhd_l2cap_connection_response_t conn_resp;
    uhd_l2cap_config_request_t conf_req;
    uhd_l2cap_disconnection_message_t disc;
    uhd_sdp_response_t sdp;
    uhd_rfcomm_message_out_t rfcomm;
  } t;
  uint8_t message_length;
  uint8_t l2capoutbuf[38];
} uhd_l2cap_queue_t;

// #define DEBUG_SDP_PDU

typedef enum sdp_pdu_type_descriptor {
  SDP_PDU_NIL = 0,
  SDP_PDU_UINT = 1,
  SDP_PDU_INT = 2,
  SDP_PDU_UUID = 3,
  SDP_PDU_STR = 4,
  SDP_PDU_BOOL = 5,
  SDP_PDU_DATASEQ = 6,
  SDP_PDU_DATAALT = 7,
  SDP_PDU_URL = 8,
  SDP_PDU_CONTAINER_END = 9,
  SDP_PDU_MAX = 9 // to 31
} sdp_pdu_type_descriptor_t;

// Stack element for recursion removal
typedef struct size_offset {
  uint16_t offset;
  uint16_t size;
} size_offset_t;

// Each type descriptor contains a size index which describes
// whether there is a 0, 1, 2 or 4-byte length specifier 
// which follows the descriptor, that encodes the length of
// the data element. This list identifies valid descriptor sizes.
struct sdp_data_element {

  public:
    sdp_data_element(uint32_t limit = 0) { reset(limit); }

  public:
    // Whenever a type size specification value is zero (indicating an
    // incomplete data element), the next two values are used to restrict the 
    // parse input range.  The missing type specification value is inserted
    // into the partial data element after the tokenizer encounters a
    // reserved token "SDP_CloseContainer" 
    size_offset_t container_sizes[8];
    uint16_t parse_limit;
    uint16_t working_offset;
    uint16_t complete:1; 
    uint16_t parse_depth:3; // Permit up to 6 levels of DE nesting
    uint16_t reserved:15; // Reserved 
    uint8_t type_descriptor;
    uint8_t type_size;
    uint32_t data_length;
    uint32_t fragment_length;
    uint8_t *data;
    uint8_t *fragment;

  public:
    void dump_stack();
    void reset(uint32_t limit = 0) {
      for ( uint8_t i = 0 ; i < 8 ; i++ ) {
        container_sizes[i].size = 0;
        container_sizes[i].offset = 0;
      }
      parse_limit = limit;
      working_offset = 0;
      complete = 0;
      parse_depth = 0; 
      data_length = 0;
      fragment_length = 0;
    };
    size_offset_t & get_typedesc0() {
      return container_sizes[0];
    };
    bool update_typespec(uint8_t * pdu, size_offset_t * stack_entry); 
};

typedef struct sdp_data_element sdp_data_element_t;

class SdpParser {
  
  public:

    bool data_element_sequence_tokenizer( sdp_data_element_t * sdp, uint8_t **pdu_token );
    uint32_t prepare_sdp_pdu_message( uint8_t * dest, const uint8_t * src, uint32_t size );
};

class TransferQueue {

  // The queue is implemented as a fixed-length FIFO.
  
  // This element can be used to uniquely identify an instance of a queue element.
  // It's value is copied to the header of newly-inserted queue elements, and
  // is incremented with each insertion.
  uint32_t queue_element_index;
  uint8_t current_position;
  uint8_t last_added;
  uint8_t available;
  uhd_l2cap_queue_t  queue[L2CAP_MESSAGEQUEUE_ITEMS];

  public:

    TransferQueue() { transfer_queue_flush(); }
    virtual ~TransferQueue() {}

    virtual bool consume_queue() = 0;
    virtual bool enqueue_l2cap_information_request(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t infoTypeLow, uint8_t infoTypeHigh) = 0;
    virtual bool enqueue_sendRfcomm(uint8_t bt_service_index, uint16_t handle, uint8_t *rfcomm_scid, uint8_t channel, uint8_t direction, uint8_t CR, uint8_t channelType, uint8_t pfBit, uint8_t* data, uint8_t length, uint32_t callback_data = 0) = 0;
    virtual bool enqueue_sendRfcommCredit(uint8_t bt_service_index, uint16_t handle, uint8_t *rfcomm_scid, uint8_t channel, uint8_t direction, uint8_t CR, uint8_t channelType, uint8_t pfBit, uint8_t credit, uint32_t callback_data = 0) = 0;

    bool transfer_queue_flush() {
      queue_element_index = 0;
      current_position = 0;
      last_added = 0;
      available = L2CAP_MESSAGEQUEUE_ITEMS;
    }

    bool transfer_queue_empty() {
      return available == L2CAP_MESSAGEQUEUE_ITEMS;
    }
    bool transfer_queue_full() {
      return available == 0;
    }

  protected:

    // Advances FIFO position and returns known currently usable entry 
    uhd_l2cap_queue_t & tq_entry(l2cap_queue_item_t t = L2CAP_QUEUE_NOOP)
    {/*{{{*/
      uint8_t e = current_position;
      current_position = (current_position + 1) % L2CAP_MESSAGEQUEUE_ITEMS;
      available--;
      memset(&(queue[e]),0, sizeof(uhd_l2cap_queue_t));
      queue[e].id = queue_element_index++;
      queue[e].type = t;
#ifdef DEBUG_TRANSFER_QUEUE
      printf("+ <%u> %u/%u T 0x%02.2X\r\n",available, current_position, L2CAP_MESSAGEQUEUE_ITEMS);
#endif
      return queue[e];
    }/*}}}*/

    l2cap_queue_item_t tq_type() {
      return queue[last_added].type;
    }

    uhd_l2cap_queue_t & tq_at(uint8_t index) {
      return queue[index];
    }

    // Return oldest entry from FIFO, store it to a memory buffer provided by caller.
    bool tq_next(uhd_l2cap_queue_t * user_buffer)
    {/*{{{*/
      uint8_t e = last_added;
      if ( transfer_queue_empty() ) return false;
      last_added = (last_added + 1) % L2CAP_MESSAGEQUEUE_ITEMS;
      available++;
      queue[e].queue_index = e;
      memmove(user_buffer, &(queue[e]), sizeof(uhd_l2cap_queue_t));
#ifdef DEBUG_TRANSFER_QUEUE
      printf("- <%u> %u/%u T 0x%02.2X \r\n",available, current_position, L2CAP_MESSAGEQUEUE_ITEMS, user_buffer->type);
#endif
      return true;
    }/*}}}*/

    // Store an L2CAP message request
    virtual bool enqueue_l2cap_message_initiator(uint8_t btd_service, l2cap_queue_item_t type, uint16_t handle, uint8_t rxid, uint8_t infoTypeLow, uint8_t infoTypeHigh ) = 0;
    virtual bool enqueue_l2cap_message(uint8_t btd_service, l2cap_queue_item_t type, uint8_t *scid, uint8_t transactionIDLow, uint8_t transactionIDHigh, uint32_t callback_payload = 0 ) = 0;
    virtual bool enqueue_l2cap_connection_response(uint8_t btd_service, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint8_t result, uint32_t callback_payload = 0) = 0;
    virtual bool enqueue_l2cap_config_request(uint8_t btd_service, uint16_t handle, uint8_t rxid, uint8_t* scid, uint32_t callback_payload = 0) = 0;
    virtual bool enqueue_l2cap_config_response(uint8_t btd_service, uint16_t handle, uint8_t rxid, uint8_t* scid, uint32_t callback_payload = 0) = 0;
    virtual bool enqueue_l2cap_disconnection_request(uint8_t btd_service, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint32_t callback_payload = 0) = 0;
    virtual bool enqueue_l2cap_disconnection_response(uint8_t btd_service, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint32_t callback_payload = 0) = 0;

};

/** All Bluetooth services should include this class. */
class BluetoothService {
  public:
    /**
     * Used to pass acldata to the Bluetooth service.
     * @param ACLData Pointer to the incoming acldata.
     */
    virtual void ACLData(uint8_t *l2capinbuf) = 0;
    /** Used to run the different state machines in the Bluetooth service. */
    virtual bool Run(uint8_t *l2capinbuf) = 0;
    /** Used to reset the Bluetooth service. */
    virtual void Reset() = 0;
    /** Update internal state of class instance */
    virtual void StateChange(uint32_t) = 0;
    /** Used to disconnect both the L2CAP Channel and the HCI Connection for the Bluetooth service. */
    virtual void disconnect() = 0;

  public:

    virtual uint8_t getServiceId() = 0;
};

class BTD;

typedef void (BTD::*HCI_State_Setter)();

// ASF UHC callbacks in the global namespace, which are invoked
// in an interrupt-handler context.

/*
 * \brief Updates the internal state of a single SPP instance to reflect that an RFCOMM protocol flow control token has been sent
 */
extern void RFCOMM_PROTOCOL_CREDIT_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data);

extern void RFCOMM_PROTOCOL_MESSAGE_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data);

/**
 * The Bluetooth Dongle class will take care of all the USBHost communication
 * and then pass the data to the BluetoothService classes.
 */
class BTD : public USBDeviceConfig, public TransferQueue, SdpParser {/*{{{*/

  friend void HCI_Command_(usb_add_t add,
		uhd_trans_status_t status,
		uint16_t payload_trans); 

  friend void HCI_event_task_(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data);

  friend void ACL_event_task_(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data); 

  friend void L2CAP_Command_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data); 

  friend void RFCOMM_PROTOCOL_CREDIT_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data);

  friend void RFCOMM_PROTOCOL_MESSAGE_handler(usb_add_t add,
    usb_ep_t ep,
    uhd_trans_status_t status,
    iram_size_t nb_transferred,
    uint8_t callback_data);

  friend class SPP;

  public:
    /**
     * Constructor for the BTD class.
     * @param  p   Pointer to USBHost class instance.
     */
    BTD(USBHost *p, const char* name = "Arduino Due SPP", const char* pin = "0000");

    virtual void setName(const char *name);

    virtual void setPin(const char *pin);

    /** @name USBDeviceConfig implementation */
    /**
     * Address assignment and basic initilization is done here.
     * @param  parent   Hub number.
     * @param  port     Port number on the hub.
     * @param  lowspeed Speed of the device.
     * @return          0 on success.
     */
    virtual uint32_t ConfigureDevice(uint32_t parent, uint32_t port, uint32_t lowspeed);
    /**
     * Initialize the Bluetooth dongle.
     * @param  parent   Hub number.
     * @param  port     Port number on the hub.
     * @param  lowspeed Speed of the device.
     * @return          0 on success.
     */
    virtual uint32_t Init(uint32_t parent, uint32_t port, uint32_t lowspeed);
    /**
     * Release the USB device.
     * @return 0 on success.
     */
    virtual uint32_t Release();
    /**
     * Poll the USB Input endpoins and run the state machines.
     * @return 0 on success.
     */
    virtual uint32_t Poll();

    /**
     * Get the device address.
     * @return The device address.
     */
    virtual uint32_t GetAddress() {
      return bAddress;
    };

    /**
     * Used to check if the dongle has been initialized.
     * @return True if it's ready.
     */
    virtual bool isReady() {
      return bPollEnable;
    };

    /** Disconnects both the L2CAP Channel and the HCI Connection for all Bluetooth services. */
    void disconnect() {
      for(uint8_t i = 0; i < BTD_NUMSERVICES; i++)
        if(!(NULL == btService[i])) {
          btService[i]->disconnect();
        }
    };

    /**
     * Register bluetooth dongle members/services.
     * @param  pService Pointer to BluetoothService class instance.
     * @return          The serice ID on succes or -1 on fail.
     */
    int8_t registerServiceClass(BluetoothService *pService) {
      for(uint8_t i = 0; i < BTD_NUMSERVICES; i++) {
        if(NULL == btService[i]) {
          btService[i] = pService;
          return i; // Return ID
        }
      }
      return -1; // ErrorregisterServiceClass
    };

    /** @name HCI Commands */
    /**
     * Used to send a HCI Command.
     * @param data   Data to send.
     * @param nbytes Number of bytes to send.
     */
    bool HCI_Command(uint8_t* data, uint16_t send_length, uint16_t read_length = 1);
    /** Read supported commands */
    void hci_read_local_supported_commands();
    /** Read supported features */
    void hci_read_local_supported_features();
    /** Set HCI default event mask */
    void hci_set_default_event_mask();
    /** Clear/Set Event Filter (OGF = 3, OCF = 5) */
    void hci_clear_event_filter();
    void hci_set_event_filter(uint8_t filter_type, uint8_t filter_cond = 0, uint8_t param_len = 0, uint8_t * filter_params = NULL);
    void hci_set_default_event_filter();
    /** Reset the Bluetooth dongle. */
    void hci_reset();
    /** Read the Bluetooth address of the dongle. */
    void hci_read_bdaddr();
    /** Read the HCI Version of the Bluetooth dongle. */
    void hci_read_local_version_information();
    /** Set ACL controller-to-Arduino flow control */
    void hci_set_controller_to_host_flow_control();
    /** Set host buffer sizes for HCI ACL and synchronous data packets */
    void hci_set_host_buffer_size();
    /** Get host buffer sizes for HCI ACL and synchronous data packets */
    void hci_read_host_buffer_size();
    /** */
    void hci_host_number_of_packets_completed(uint16_t packets);
    /** Set flow parameters for connection handle */
    void hci_set_flow_specification();
    /**
     * Set the local name of the Bluetooth dongle.
     * @param name Desired name.
     */
    void hci_set_local_name(const char* name);
    /** Enable SSP */
    void hci_write_simple_pairing_mode(bool v);
    /** Enable visibility to other Bluetooth devices. */
    void hci_write_scan_enable();
    /** Set page scan timeout */
    void hci_write_page_timeout();
    /** Set connection accept timeout */
    void hci_write_connection_accept_timeout();
    /** Disable visibility to other Bluetooth devices. */
    void hci_write_scan_disable();
    /** Read the remote devices name. */
    void hci_remote_name();
    /** Write Extended Inquiry Response (CRITICAL - Publishes Service UUIDs) */
    void hci_write_extended_inquiry_response();
    /** Accept the connection with the Bluetooth device. */
    void hci_accept_connection();
    /** Send confirmation */
    void hci_user_confirmation_request_reply();
    /**
     * Disconnect the HCI connection.
     * @param handle The HCI Handle for the connection.
     */
    void hci_disconnect(uint16_t handle);
    /**
     * Respond with the pin for the connection.
     * The pin is automatically set for the Wii library,
     * but can be customized for the SPP library.
     */
    void hci_pin_code_request_reply();
    /** Response when no pin was set. */
    void hci_pin_code_negative_request_reply();
    /**
     * Command is used to reply to a Link Key Request event from the BR/EDR Controller
     * if the Host does not have a stored Link Key for the connection.
     */
    void hci_link_key_request_negative_reply();
    void hci_link_key_request_reply();
    /** Change connection packet type handling for current connection handle */
    void hci_change_connection_packet_type();
    /** Set/clear authentication enable flag. */
    void hci_write_authentication_enable(bool);
    /** Used to try to authenticate with the remote device. */
    void hci_authentication_request();
    /** Start a HCI inquiry. */
    void hci_inquiry();
    /** Cancel a HCI inquiry. */
    void hci_inquiry_cancel();
    /** Connect to a device. */
    void hci_connect();
    /** Used to a set the class of the device. */
    void hci_write_class_of_device();

    /** SSP implementation methods: OOB data read, IO capability response, etc. */
    void hci_io_capability_response();
    void hci_write_pin_type(bool variable);
    void hci_read_local_oob_data();
    void hci_oob_data_request_reply();
    void hci_link_key_notification();
    void hci_delete_stored_link_key(uint8_t * bd_addr);
    void hci_read_stored_link_key(uint8_t * bd_addr);
    void hci_write_stored_link_key();

    /** Link Encryption methods */
    void hci_master_link_key(); // 7.1.18 Master Link Key Command
    void hci_change_connection_link_key(); // 7.1.17 Change Connection Link Key Command
    void hci_set_connection_encryption(); // 7.1.16 Set Connection Encryption Command

    /** Remote query methods - Link Control  */
    void hci_read_remote_supported_features(); // 7.1.21 Read Remote Supported Features Command
    void hci_read_remote_extended_features(); // 7.1.22 Read Remote Extended Features Command
    void hci_read_remote_version_information(); // 7.1.23 Read Remote Version Information Command

    /** Debug / diagnostics */
    void hci_dump_event();
    void hci_dump_command();
    void hci_dump_completed_packets_event(uint8_t *event_data);


    bool gotNak() {
      return pUsb->GetNak();
    }

    /** L2CAP Commands */

    /*
     * \brief Implement deferred command queue
     *
     * The ASF (Atmel Software Framework) uses callbacks to notify 
     * consumers of USB subsystem events, and implements low-level pipe
     * transfers.  Transfers are no longer guaranteed to be complete at
     * point of call, so the state machine methods must leave interrupt context,
     * and defer execution of USB I/O until the callback infrastructure notifies
     * upper layer code that it can proceed to update higher-level state and
     * execute a [possibly empty] sequence of command events.
     *
     * Instead of executing USB transfers (which no longer block) in lock step,
     * we push commands onto a ring queue, and defer execution of these commands
     * until we leave interrupt context.  The following methods implement queue
     * management and transfer dispatch for class BluetoothService.
     */

    bool enqueue_l2cap_message_initiator(uint8_t bt_service_index, l2cap_queue_item_t type, uint16_t handle, uint8_t rxid, uint8_t infoTypeLow, uint8_t infoTypeHigh );
    bool enqueue_l2cap_message(uint8_t bt_service_index, l2cap_queue_item_t type, uint8_t *scid, uint8_t transactionIDLow, uint8_t transactionIDHigh, uint32_t callback_payload = 0 );
    bool enqueue_l2cap_information_request(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t infoTypeLow, uint8_t infoTypeHigh)
    {/*{{{*/
      return enqueue_l2cap_message_initiator(bt_service_index, L2CAP_QUEUE_INFORMATION_REQUEST, handle, rxid, infoTypeLow, infoTypeHigh);
    }/*}}}*/
    bool enqueue_l2cap_connection_response(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint8_t result, uint32_t callback_payload = 0);
    bool enqueue_l2cap_config_request(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* scid, uint32_t callback_payload = 0);
    bool enqueue_l2cap_config_response(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* scid, uint32_t callback_payload = 0);
    bool enqueue_l2cap_disconnection_request(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint32_t callback_payload = 0);
    bool enqueue_l2cap_disconnection_response(uint8_t bt_service_index, uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint32_t callback_payload = 0);
    bool enqueue_sendRfcomm(uint8_t bt_service_index, uint16_t handle, uint8_t *rfcomm_scid, uint8_t channel, uint8_t direction, uint8_t CR, uint8_t channelType, uint8_t pfBit, uint8_t* data, uint8_t length, uint32_t callback_data = 0);
    bool enqueue_sendRfcommCredit(uint8_t bt_service_index, uint16_t handle, uint8_t *rfcomm_scid, uint8_t channel, uint8_t direction, uint8_t CR, uint8_t channelType, uint8_t pfBit, uint8_t credit, uint32_t callback_data = 0);
    bool consume_queue();
    uint8_t get_service_index(uint8_t callback_data);
    uint32_t get_instance_payload(uint8_t queue_index);

    void invoke_callback_state_update(usb_add_t add,
        usb_ep_t ep,
        uhd_trans_status_t status,
        iram_size_t nb_transferred,
        uint8_t callback_data);

    /**
     * Used to send L2CAP Commands.
     * @param handle      HCI Handle.
     * @param data        Data to send.
     * @param nbytes      Number of bytes to send.
     * @param channelLow,channelHigh  Low and high byte of channel to send to.
     * @param callback    Method invoked by ASF USB pipe processing callback
     * @param payload     Five-bit payload data passed to USB pipe-processing callback
     * If argument is omitted then the Standard L2CAP header: Channel ID (0x01) for ACL-U will be used.
     */
    bool L2CAP_Command(uint16_t handle, uint8_t* data, uint8_t nbytes, uint8_t channelLow = 0x01, uint8_t channelHigh = 0x00, uhd_callback_trans_t callback = NULL, uint8_t callback_data = 0 );
    /**
     * L2CAP Connection Request.
     * @param handle HCI handle.
     * @param rxid   Identifier.
     * @param scid   Source Channel Identifier.
     * @param psm    Protocol/Service Multiplexer - see: https://www.bluetooth.org/Technical/AssignedNumbers/logical_link.htm.
     */
    void l2cap_connection_request(uint16_t handle, uint8_t rxid, uint8_t* scid, uint16_t psm);
    /**
     * L2CAP Config Response.
     * @param handle HCI Handle.
     * @param rxid   Identifier.
     * @param scid   Source Channel Identifier.
     */
    void l2cap_config_response(uint16_t handle, uint8_t rxid, uint8_t* scid);
    /**
     * L2CAP Disconnection Request.
     * @param handle HCI Handle.
     * @param rxid   Identifier.
     * @param dcid   Device Channel Identifier.
     * @param scid   Source Channel Identifier.
     */
    void l2cap_disconnection_request(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid);

    /* RFCOMM Commands */
    void RFCOMM_Command(uint8_t *rfcomm_scid, uint8_t* data, uint8_t nbytes);
    uint8_t calcFcs(uint8_t *data);
    uint8_t __crc(uint8_t* data);

    /////////////////////////////////////////////////////////////////////////

    /** Use this to see if it is waiting for a incoming connection. */
    bool connectionWait;
    /** This is used by the service to know when to store the device information. */
    bool l2capConnectionClaimed;
    /** This is used by the SPP library to claim the current SDP incoming request. */
    bool sdpConnectionClaimed;
    /** This is used by the SPP library to claim the current RFCOMM incoming request. */
    bool rfcommConnectionClaimed;

    /** Link key and type for pairing with a single peer */
    // See: EV_LINK_KEY_NOTIFICATION, hci_link_key_notification
    uint8_t link_key_type;
    uint8_t link_key[16];

    /** The name you wish to make the dongle show up as. It is set automatically by the SPP library. */
    char btdName[250];
    /** The pin you wish to make the dongle use for authentication. It is set automatically by the SPP library. */
    char btdPin[17]; // NUL-terminated, up to 128 bits long

    /** The Bluetooth dongle's Bluetooth address. */
    uint8_t my_bdaddr[6];
    /** Last incoming device's Bluetooth address. */
    uint8_t disc_bdaddr[6];
    /** OOB Pairing Hash C, per Core Vol. 4, Part 2.E.7.3.60 */
    uint8_t oob_pairing_hash[16];
    /** OOB Pairing Randomizer R, per Core Vol. 4, Part 2.E.7.3.60 */
    uint8_t oob_pairing_randomizer[16];
    /** First 30 chars of last remote name. */
    uint8_t remote_name[30];

    uint8_t l2capinbuf[L2CAP_IN_BUFFERSIZE]; //General purpose buffer for l2cap in data
    uint8_t l2capoutbuf[L2CAP_OUT_BUFFERSIZE]; // General purpose buffer for l2cap out data
    uint8_t local_features_bitmask[8];

    /**
     * The supported HCI Version read from the Bluetooth dongle.
     * Used by the PS3BT library to check the HCI Version of the Bluetooth dongle,
     * it should be at least 3 to work properly with the library.
     */
    uint8_t hci_version;

    /**
     * Read the poll interval taken from the endpoint descriptors.
     * @return The poll interval in ms.
     */
    uint8_t readPollInterval() {
      return pollInterval;
    };

  protected:
    /** Pointer to USB class instance. */
    USBHost *pUsb;
    /** Pointer to last-invoked HCI state handler */
    HCI_State_Setter hci_last_invoked;
    /** Device address. */
    uint32_t bAddress;
    /** Number of configurations */
    uint8_t bNumConfigurations;
    /** Maximum packet size */
    uint8_t maxPktSize;
    /** Endpoint info structure. */
    EpInfo epInfo[BTD_MAX_ENDPOINTS];

    /** Configuration number. */
    uint32_t bConfNum;
    /** Total number of endpoints in the configuration. */
    uint32_t bNumEP;
    /** Next poll time based on poll interval taken from the USB descriptor. */
    uint32_t qNextPollTime;
    /** SAM internal tick count */
    uint32_t ticks;

  public:
    /** Bluetooth dongle control endpoint. */
    static const uint32_t BTD_CONTROL_PIPE;
    /** HCI event endpoint index. */
    static const uint32_t BTD_EVENT_PIPE;
    /** ACL In endpoint index. */
    static const uint32_t BTD_DATAIN_PIPE;
    /** ACL Out endpoint index. */
    static const uint32_t BTD_DATAOUT_PIPE;

  protected:
    /**
     * Used to print the USB Endpoint Descriptor.
     * @param ep_ptr Pointer to USB Endpoint Descriptor.
     */
    void PrintEndpointDescriptor(const USB_ENDPOINT_DESCRIPTOR* ep_ptr);

  private:

    void resetBtServices();
    void clearAllVariables(); // Set all variables, endpoint structs etc. to default values

    BluetoothService* btService[BTD_NUMSERVICES];
    uint16_t PID, VID; // PID and VID of device connected

    bool bPollEnable;
    bool bSspCapable;
    bool bEirCapable;
    bool bConnected; // HCI handle available, incoming connection accepted
    uint8_t pollInterval;

    /* ASF callback infrastructure support */

    /* Variables used by high level HCI task */
    uint8_t hci_state; //current state of bluetooth hci connection
    uint32_t hci_counter; // counter used for bluetooth hci reset loops
    uint32_t hci_num_reset_loops; // this value indicate how many times it should read before trying to reset
    uint32_t hci_event_flag; // hci flags of received bluetooth events
    uint8_t inquiry_counter;
    uint8_t completed_packet_events; 
    /** HCI handle for the last connection. */
    uint16_t hci_handle;

    uint8_t hcibuf[HCI_BUFFERSIZE]; //General purpose buffer for hci data
    uint8_t hcicmd[HCI_CMDBUFSIZE]; //General purpose buffer for hci data

    /* State machines */
    bool HCI_event_task(uhd_trans_status_t status = UHD_TRANS_NOERROR); // Poll the HCI event pipe
    void HCI_state_update(uhd_trans_status_t status); // HCI state updates
    void HCI_task(); // HCI state machine

    bool ACL_event_task();
    void ACL_state_update(uhd_trans_status_t status = UHD_TRANS_NOERROR);

};/*}}}*/

#endif
